/***************************************************************
 * Name:      EDID_CEA.h
 * Purpose:   EDID Extension structures: CEA
 * Author:    Tomasz Pawlak (tomasz.pawlak@wp.eu)
 * Created:   2014-03-28
 * Copyright: Tomasz Pawlak (C) 2014-2018
 * License:   GPLv3
 **************************************************************/

#ifndef EDID_CEA_H
#define EDID_CEA_H 1

#include <stdint.h>

typedef unsigned int uint;

enum { //Block Header : type tags
   BTYPE_ADB = 1, //Audio Data Block
   BTYPE_VDB = 2, //Video Data Block
   BTYPE_VSD = 3, //Vendor Specific Data Block
   BTYPE_SAB = 4  //Speaker Allocation Data Block
};

//Block Header
typedef struct __attribute__ ((packed)) { //tag_s
   uint blk_len  :5;    //bit 4-0: Total number of bytes in the block exluding this byte
   uint blk_type :3;    /*bit 7-5: 1=audio, 2=video, 3=vendor-specific, 4=speaker allocation,
                                   >4 reserved */
} tag_t;

typedef union __attribute__ ((packed)) { //bhdr_s
   uint8_t bval;
   tag_t   tag;
} bhdr_t;

//ADB: Audio Data Block ->
//SAD: Short Audio Descriptor
 /*Audio format code
1 = Linear Pulse Code Modulation (LPCM)
2 = AC-3
3 = MPEG1 (Layers 1 and 2)
4 = MP3
5 = MPEG2
6 = AAC
7 = DTS
8 = ATRAC
0, 15: Reserved
9 = One-bit audio aka SACD
10 = DD+
11 = DTS-HD
12 = MLP/Dolby TrueHD
13 = DST Audio
14 = Microsoft WMA Pro
*/

typedef struct __attribute__ ((packed)) { //sad0_s
   // byte 0;            // SAD byte 0: format and number of channels
   uint  num_ch    :3;   //bit 2-0: Num of channels -1 (7->8 channels, 1-> 2ch)
   uint  audio_fmt :4;   //bit 6-3: Audio format code
   uint  resvd0    :1;   //bit 7:   reserved.
} sad0_t;

typedef struct __attribute__ ((packed)) { //sad1_s
   // byte 0;            // SAD byte 1: sampling frequencies supported
   uint  sf32k     :1;   //bit 0: sampling frequency 32kHz supported
   uint  sf44k     :1;   //bit 1: sampling frequency 44kHz supported
   uint  sf48k     :1;   //bit 2: sampling frequency 48kHz supported
   uint  sf88k     :1;   //bit 3: sampling frequency 88kHz supported
   uint  sf96k     :1;   //bit 4: sampling frequency 96kHz supported
   uint  sf176k    :1;   //bit 5: sampling frequency 176kHz supported
   uint  sf192k    :1;   //bit 6: sampling frequency 192kHz supported
   uint  resvd1    :1;   //bit 7: reserved.
} sad1_t;

typedef union __attribute__ ((packed)) { //sad2_s
   struct __attribute__ ((packed)) { //SAD3 byte0 -> LPCM: bits 7:3 are reserved and 2:0 bits define bit depth
      uint sample16b   :1; //LPCM: 16bit sample
      uint sample20b   :1; //LPCM: 20bit sample
      uint sample24b   :1; //LPCM: 24bit sample
      uint resvd       :5;
   } lpcm;
   uint8_t bitrate    ; //bitrate dib by 8kbit/s
} sad2_t;

//SAD: base struct
typedef struct __attribute__ ((packed)) { //sad_s
   sad0_t sad0; //1: SAD byte 0: format and number of channels
   sad1_t sad1; //2: SAD byte 1: sampling frequencies supported
   sad2_t sad2; //3: SAD byte 2: bitrate
} sad_t;

//VDB: Video Data Block ->
//SVD: Short Video Descriptor base struct
typedef struct __attribute__ ((packed)) { //svd_s
   uint  vidmd_idx :7;  //1: bit 6-0: video mode idx: 0, 64-127 reserved
   uint  native    :1;  //1: bit 7  : native mode @ idx
} svd_t;

//VSD: Vendor Specific Data Block
//VSD: Sink supported features
typedef struct __attribute__ ((packed)) { //vsd_feat_s
   uint    dvi_dual :1;  //bit0 : DVI Dual Link Operation
   uint    resvd0   :1;  //bit1 : reserved
   uint    resvd1   :1;  //bit2 : reserved
   uint    dc_y444  :1;  //bit3 : DC_Y444  (supports 4:4:4 in deep color modes)
   uint    dc_30bit :1;  //bit4 : DC_30bit (supports 10-bit-per-channel deep color)
   uint    dc_36bit :1;  //bit5 : DC_36bit (supports 10-bit-per-channel deep color)
   uint    dc_48bit :1;  //bit6 : DC_48bit (supports 10-bit-per-channel deep color)
   uint    ai_supp  :1;  //bit7 : Supports_AI (needs info from ACP or ISRC packets)
} vsd_feat_t;

//VSD: Latency
typedef struct __attribute__ ((packed)) { //vsd_ltncy_s
   uint    resvd    :6;  //bit0-5 : reserved
   uint    i_ltncy  :1;  /*bit6 : i_latency_fields (set if interlaced latency fields are present; if set
                                  four latency fields will be present, 0 if bit 7 is 0) */
   uint    ltncy_f  :1;  //bit7 : latency_fields (set if latency fields are present)
} vsd_ltncy_t;

//VSD:
typedef struct __attribute__ ((packed)) { //vsd_s
   bhdr_t      tag;        //0:
   uint8_t     ieee_id[3]; //1-3: IEEE Registration Id LE, 00-0C-03 for HDMI Licensing, LLC
   uint16_t    src_phy;    //4-5: Source Physical Address (section 8.7 of HDMI 1.3a)
   vsd_feat_t  sink_feat;  //6: sink supported features
   uint8_t     max_tmds;   //7: (opt) If non-zero (Max_TMDS_Frequency / 5mhz)
   vsd_ltncy_t ltncy_hdr;  //8: (opt) latency fields indicators

   uint8_t     vid_lat;    //9: (opt) Video Latency value=1+ms/2 with a max of 251 meaning 500ms
   uint8_t     aud_lat;    /* 10: (opt) Audio Latency value=1+ms/2 with a max of 251 meaning 500ms
                                  (video delay for progressive sources */
   uint8_t     vid_ilat;   //11: (opt) Interlaced Video Latency value=1+ms/2 with a max of 251 meaning 500ms
   uint8_t     aud_ilat;   /* 12: (opt) Interlaced Audio Latency value=1+ms/2 with a max of 251 meaning 500ms
                                   (video delay for interlaced sources) */
   //Additional bytes may be present, but the HDMI spec says they shall be zero.
} vsd_t;

//SAB: Speaker Allocation Block
typedef struct __attribute__ ((packed)) { //sab0_s
   uint    front_lr   :1;  //bit0: Front Left / Right present
   uint    lfe        :1;  //bit1: LFE present
   uint    front_ct   :1;  //bit2: Front Center present
   uint    rear_lr    :1;  //bit3: Rear Left / Right present
   uint    rear_ct    :1;  //bit4: Rear Center present
   uint    frontct_lr :1;  //bit5: Front Left Center / Right Center present
   uint    rearct_lr  :1;  //bit6: Rear Left Center / Right Center present
   uint    resvd      :1;  //bit7: reserved (0)
} sab0_t;

typedef struct __attribute__ ((packed)) { //sab_s
   bhdr_t  tag;        //0:
   sab0_t  spk_flags;  //1 speaker presence flags
   uint8_t resvdbt1;   //2: reserved
   uint8_t resvdbt2;   //3: reserved
} sab_t;

// CEA/EIA-861 header
//extension block info
typedef struct __attribute__ ((packed)) { //infoblk_s
   uint    num_dtd     :4;  //bit 3-0: total number of the DTDs included in this block
   uint    ycbcr422    :1;  //bit4:    1 = supports YCbCr 4:2:2
   uint    ycbcr444    :1;  //bit5:    1 = supports YCbCr 4:4:4
   uint    basic_audio :1;  //bit6:    1 = supports basic audio
   uint    underscan   :1;  //bit7:    1 = supports underscan
} infoblk_t;

//CEA header
typedef struct __attribute__ ((packed)) { //cea_hdr_s
   uint8_t    ext_tag;   //0 : extension id 0x02 for CEA Timing Extension (CEA-EDID)
   uint8_t    rev;       //1 : extension block revision
   uint8_t    dtd_offs;  //2 : offset to first DTD, if dtd_offs=0 -> No DTD sections
   infoblk_t  info_blk;  //3 : Extension block info: number of DTDs present, other Version 2+ information
   uint8_t    dta_start; //4 : First data byte
/* Content: (start at 4th byte)
if dtd_offs!=4 DBC starts here (Data Block Collection), then
at dtd_offs 18-byte DTD strings continue for an unspecified
length (modulo 18) until a "00 00" is as the first bytes of a prospective DTD.
At this point, the DTDs are known to be complete, and the start address of the "00 00"
can be considered to be "XX".
then:
"XX"-126: Post-DTD padding.  Should be populated with 00h
127: Checksum - This byte must be programmed such that the sum of all 128 bytes equals 00h.
if dtd_offs!=4 DBC starts here (Data block Collection)
*/
} cea_hdr_t;

#endif /* EDID_CEA_H */
