/***************************************************************
 * Name:      EDID_CEA_class.cpp
 * Purpose:   EDID CEA-861 extension
 * Author:    Tomasz Pawlak (tomasz.pawlak@wp.eu)
 * Created:   2014-05-08
 * Copyright: Tomasz Pawlak (C) 2014-2018
 * License:   GPLv3
 **************************************************************/

#include "debug.h"
#include "scopes.h"
#ifndef scEDID_CEA
   #error "EDID_class.h: missing scope definition"
#endif
#define RCD_SCOPE scEDID_CEA
#include "returncode.h"

#include <stddef.h>

#define EDID_CEA_CLASS
#include "EDID_class.h"
#undef  EDID_CEA_CLASS

//CHD: CEA Header
const wxString cea_hdr_cl::Name = _("CHD: CEA-861 header");
const wxString cea_hdr_cl::Desc = _("CEA-861 header");
const size_t cea_hdr_cl::fcount = 9;


const edi_field_t cea_hdr_cl::fields[] = {
   {&EDID_cl::ByteVal, NULL, offsetof(cea_hdr_t, ext_tag), 0, 1, EF_HEX|EF_RD, 0, 0xFF,
   "Extension tag", "Extension type: 0x02 for CEA-861" },
   {&EDID_cl::ByteVal, NULL, offsetof(cea_hdr_t, rev), 0, 1, EF_INT|EF_RD, 0, 0xFF, "Revision",
   "Extension revision" },
   {&EDID_cl::ByteVal, NULL, offsetof(cea_hdr_t, dtd_offs), 0, 1, EF_HEX|EF_RD, 0, 0xFF,
   "DTD offset", "Offset to first DTD, if dtd_offs=0 -> No DTD sections" },
   {&EDID_cl::BitF8Val, NULL, offsetof(cea_hdr_t, info_blk), 0, 4, EF_BFLD|EF_INT|EF_RD, 0, 0x0F,
   "num_dtd", "Byte3, bits 3-0: total number of the DTD sections in this extension." },
   {&EDID_cl::BitVal, NULL, offsetof(cea_hdr_t, info_blk), 4, 1, EF_BIT, 0, 1, "YCbCr 4:2:2",
   "1 = supported." },
   {&EDID_cl::BitVal, NULL, offsetof(cea_hdr_t, info_blk), 5, 1, EF_BIT, 0, 1, "YCbCr 4:4:4",
   "1 = supported." },
   {&EDID_cl::BitVal, NULL, offsetof(cea_hdr_t, info_blk), 6, 1, EF_BIT, 0, 1, "Basic Audio",
   "1 = supports basic audio." },
   {&EDID_cl::BitVal, NULL, offsetof(cea_hdr_t, info_blk), 7, 1, EF_BIT, 0, 1, "Underscan",
   "1 = supports underscan." },
   {&EDID_cl::ByteVal, NULL, 127, 0, 1, EF_HEX|EF_RD, 0, 0xFF, "chksum",
   "Block checksum: sum of all bytes mod 256 must equal zero." }
};

rcode cea_hdr_cl::init(uint8_t* inst, uint32_t orflags) {
   rcode retU;

   if (inst == NULL) {
      RCD_SET_FAULT(retU);
      return retU;
   }
   type_id   = ID_CHD;
   GroupName = &Name;
   GroupDesc = &Desc;

   retU = init_fields(&fields[0], inst, fcount, orflags);
   return retU;
}

//Block Header Describtions
static const char CEA_BlkHdr_dscF0[] = "Total number of bytes in the block excluding header byte";
static const char CEA_BlkHdr_dscF1[] = "Block Type Tag: \n "
"1 = ADB: Audio Data Block\n 2 = VDB: Video Data Block\n 3 = VSD: Vendor Specific Data Block\n "
"4 = SAB: Speaker Allocation Data Block\nall other values Reserved)";

//Block Header fields
const edi_field_t CEA_BlkHdr_fields[] = {
   {&EDID_cl::BitF8Val, NULL, offsetof(bhdr_t, tag), 0, 5, EF_BFLD|EF_INT|EF_RD, 0, 0x1F, "blk_len",
   CEA_BlkHdr_dscF0 },
   {&EDID_cl::BitF8Val, NULL, offsetof(bhdr_t, tag), 5, 3, EF_BFLD|EF_INT|EF_RD, 0, 0x07, "blk_type",
    CEA_BlkHdr_dscF1 }
};

//ADB: Audio Data Block

//ADB: Audio Data Block : audio format codes
const vname_map_t adb_audiofmt_map[] = {
   {0, "Reserved"},
   {1, "LPCM"},
   {2, "AC-3"},
   {3, "MPEG1 (Layers 1 and 2)"},
   {4, "MP3"},
   {5, "MPEG2"},
   {6, "AAC"},
   {7, "DTS"},
   {8, "ATRAC"},
   {9, "SACD"},
   {10, "DD+"},
   {11, "DTS-HD"},
   {12, "MLP/Dolby TrueHD"},
   {13, "DST Audio"},
   {14, "Microsoft WMA Pro"},
   {15, "Reserved"}
};

const vmap_t ADB_audiofmt = {
   0, 16,
   adb_audiofmt_map
};

const wxString cea_adb_cl::Name = _("ADB: Audio Data Block");
const wxString cea_adb_cl::Desc = _("Audio Data Block contains one or more 3-byte "
"Short Audio Descriptors (SADs). Each SAD details audio format, channel number, "
"and bitrate/resolution capabilities of the display");
const size_t cea_adb_cl::fcount = 2;

rcode cea_adb_cl::init(uint8_t* inst, uint32_t orflags) {
   rcode retU;
   bhdr_t  bhdr;
   uint    blen;

   if (inst == NULL) {
      RCD_SET_FAULT(retU);
      return retU;
   }
   bhdr.bval = reinterpret_cast <bhdr_t*> (inst)->bval;
   //chk type
//   if (bhdr.tag.blk_type != BTYPE_ADB) {
//      RCD_SET_FAULT(retU);
//      return retU;
//   }
   //chk len
   blen = bhdr.tag.blk_len;
   if ((blen % sizeof(sad_t)) != 0) {
      RCD_SET_FAULT(retU);
      return retU;
   }
   blen      = blen / sizeof(sad_t);
   type_id   = ID_ADB;
   GroupName = &Name;
   GroupDesc = &Desc;

   RCD_SET_OK(retU);
   sad_t* pgrp_inst = reinterpret_cast <sad_t*> (inst+1);
   subgroups.Alloc(blen);
   //create and init sub-groups: SAD: Short Audio Descriptor
   for (uint idx=0; idx<blen; idx++) {
      edi_grp_cl* pgrp = new cea_sad_cl;
      if (pgrp == NULL) {
         RCD_SET_FAULT(retU);
         break;
      }
      retU = pgrp->init(reinterpret_cast <uint8_t*> (pgrp_inst), 0);
      if (! RCD_IS_OK(retU)) break;
      pgrp->setRelOffs(reinterpret_cast <uint8_t*> (pgrp_inst) - inst);
      pgrp->setAbsOffs(abs_offs + pgrp->getRelOffs() );
      subgroups.Add(pgrp);
      pgrp_inst += 1;
   }
   if (! RCD_IS_OK(retU)) {
      subgroups.Clear();
      return retU;
   }

   retU = init_fields(&CEA_BlkHdr_fields[0], inst, fcount, orflags);
   return retU;
}

//ADB: Audio Data Block ->
//SAD: Short Audio Descriptor
const wxString cea_sad_cl::Name = _("SAD: Short Audio Descriptor");
const wxString cea_sad_cl::Desc = _("SAD describes audio format, channel number, "
"and bitrate/resolution capabilities of the display");
static const char freqSuppDsc[] = "SAD byte 1, bit flag =1: sampling frequency is supported.";
const size_t cea_sad_cl::fcount = 16;

const edi_field_t cea_sad_cl::fields[] = {
   {&EDID_cl::BitF8Val, NULL, offsetof(sad_t, sad0), 0, 3, EF_BFLD|EF_INT, 0, 0xFF, "num_ch",
   "Bitfield of 3-bits, number of channels minus 1, 3-bit bitfield  (i.e. 000 = 1 channel; "
   "001 = 2 channels; 111 = 8 channels)" },
   {&EDID_cl::BitF8Val, &ADB_audiofmt, offsetof(sad_t, sad0), 3, 4, EF_BFLD|EF_INT|EF_VS, 0, 0x0F, "audio_fmt",
   "Audio format code (index):\n"
   "0, 15: Reserved\n"
   "1 = Linear Pulse Code Modulation (LPCM)\n"
   "2 = AC-3\n"
   "3 = MPEG1 (Layers 1 and 2)\n"
   "4 = MP3\n"
   "5 = MPEG2\n"
   "6 = AAC\n"
   "7 = DTS\n"
   "8 = ATRAC\n"
   "9 = SACD\n"
   "10 = DD+\n"
   "11 = DTS-HD\n"
   "12 = MLP/Dolby TrueHD\n"
   "13 = DST Audio\n"
   "14 = Microsoft WMA Pro" },
   {&EDID_cl::BitVal, NULL, offsetof(sad_t, sad0), 7, 1, EF_BIT, 0, 1, "reserved",
   "Reserved (0)." },
   {&EDID_cl::BitVal, NULL, offsetof(sad_t, sad1), 0, 1, EF_BIT, 0, 1, "sf_32kHz",
   freqSuppDsc },
   {&EDID_cl::BitVal, NULL, offsetof(sad_t, sad1), 1, 1, EF_BIT, 0, 1, "sf_44kHz",
   freqSuppDsc },
   {&EDID_cl::BitVal, NULL, offsetof(sad_t, sad1), 2, 1, EF_BIT, 0, 1, "sf_48kHz",
   freqSuppDsc },
   {&EDID_cl::BitVal, NULL, offsetof(sad_t, sad1), 3, 1, EF_BIT, 0, 1, "sf_88kHz",
   freqSuppDsc },
   {&EDID_cl::BitVal, NULL, offsetof(sad_t, sad1), 4, 1, EF_BIT, 0, 1, "sf_96kHz",
   freqSuppDsc },
   {&EDID_cl::BitVal, NULL, offsetof(sad_t, sad1), 5, 1, EF_BIT, 0, 1, "sf_176kHz",
   freqSuppDsc },
   {&EDID_cl::BitVal, NULL, offsetof(sad_t, sad1), 6, 1, EF_BIT, 0, 1, "sf_192kHz",
   freqSuppDsc },
   {&EDID_cl::BitVal, NULL, offsetof(sad_t, sad1), 7, 1, EF_BIT, 0, 1, "reserved",
   "Reserved (0)." },
   {&EDID_cl::BitVal, NULL, offsetof(sad_t, sad2), 0, 1, EF_BIT, 0, 1, "sample16b",
   "LPCM sample size in bits: 1=supported." },
   {&EDID_cl::BitVal, NULL, offsetof(sad_t, sad2), 1, 1, EF_BIT, 0, 1, "sample20b",
   "LPCM sample size in bits: 1=supported." },
   {&EDID_cl::BitVal, NULL, offsetof(sad_t, sad2), 2, 1, EF_BIT, 0, 1, "sample24b",
   "LPCM sample size in bits: 1=supported." },
   {&EDID_cl::BitVal, NULL, offsetof(sad_t, sad2), 3, 5, EF_BIT, 0, 0, "LPCM_resvd",
   "For LPCM this is reserved (0)" },
   {&EDID_cl::ByteVal, NULL, offsetof(sad_t, sad2), 0, 1, EF_INT, 0, 0xFF, "Bitrate",
   "For non-LPCM format, this is the maximum supported bitrate divided by 8 kbit/s.\n"
   "This field is overlapping LPCM fields" }
};

rcode cea_sad_cl::init(uint8_t* inst, uint32_t orflags) {
   rcode retU;

   if (inst == NULL) {
      RCD_SET_FAULT(retU);
      return retU;
   }
   type_id   = ID_SAD;
   GroupName = &Name;
   GroupDesc = &Desc;

   retU = init_fields(&fields[0], inst, fcount, orflags);
   return retU;
}

//VDB: Video Data Block : video modes
/* Notes:
Parentheses indicate instances where pixels are repeated to meet the minimum speed
requirements of the interface. For example, in the 720X240p case, the pixels on each line
are double-clocked. In the (2880)X480i case, the number of pixels on each line, and thus
the number of times that they are repeated, is variable, and is sent to the DTV monitor by
the source device.
*/

const vname_map_t vidmd_map[] = {
   {1,  "DMT0659   4:3 640x480p@59.94/60Hz"},
   {2,  "480p      4:3 720x480p@59.94/60Hz"},
   {3,  "480pH    16:9 720x480p@59.94/60Hz"},
   {4,  "720p     16:9 1280x720p@59.94/60Hz"},
   {5,  "1080i    16:9 1920x1080i@59.94/60Hz"},
   {6,  "480i      4:3 720(1440)x480i@59.94/60Hz"},
   {7,  "480iH    16:9 720(1440)x480i@59.94/60Hz"},
   {8,  "240p      4:3 720(1440)x240p@59.94/60Hz"},
   {9,  "240pH    16:9 720(1440)x240p@59.94/60Hz"},
   {10, "480i4x    4:3 (2880)x480i@59.94/60Hz"},
   {11, "480i4xH  16:9 (2880)x480i@59.94/60Hz"},
   {12, "240p4x    4:3 (2880)x240p@59.94/60Hz"},
   {13, "240p4xH  16:9 (2880)x240p@59.94/60Hz"},
   {14, "480p2x    4:3 1440x480p@59.94/60Hz"},
   {15, "480p2xH  16:9 1440x480p@59.94/60Hz"},
   {16, "1080p    16:9 1920x1080p@59.94/60Hz"},
   {17, "576p      4:3 720x576p@50Hz"},
   {18, "576pH    16:9 720x576p@50Hz"},
   {19, "720p50   16:9 1280x720p@50Hz"},
   {20, "1080i25  16:9 1920x1080i(1125 Vtotal)@50Hz"},
   {21, "576i      4:3 720(1440)x576i@50Hz"},
   {22, "576iH    16:9 720(1440)x576i@50Hz"},
   {23, "288p      4:3 720(1440)x288p@50Hz"},
   {24, "288pH    16:9 720(1440)x288p@50Hz"},
   {25, "576i4x    4:3 (2880)x576i@50Hz"},
   {26, "576i4xH  16:9 (2880)x576i@50Hz"},
   {27, "288p4x    4:3 (2880)x288p@50Hz"},
   {28, "288p4xH  16:9 (2880)x288p@50Hz"},
   {29, "576p2x    4:3 1440x576p@50Hz"},
   {30, "576p2xH  16:9 1440x576p@50Hz"},
   {31, "1080p50  16:9 1920x1080p@50Hz"},
   {32, "1080p24  16:9 1920x1080p@23.98/24Hz"},
   {33, "1080p25  16:9 1920x1080p@25Hz"},
   {34, "1080p30  16:9 1920x1080p@29.97/30Hz"},
   {35, "480p4x    4:3 (2880)x480p@59.94/60Hz"},
   {36, "480p4xH  16:9 (2880)x480p@59.94/60Hz"},
   {37, "576p4x    4:3 (2880)x576p@50Hz"},
   {38, "576p4xH  16:9 (2880)x576p@50Hz"},
   {39, "1080i25  16:9 1920x1080i(1250 VTotal)@50Hz"},
   {40, "1080i50  16:9 1920x1080i@100Hz"},
   {41, "720p100  16:9 1280x720p@100Hz"},
   {42, "576p100   4:3 720x576p@100Hz"},
   {43, "576p100H 16:9 720x576p@100Hz"},
   {44, "576i50    4:3 720(1440)x576i@100Hz"},
   {45, "576i50H  16:9 720(1440)x576i@100Hz"},
   {46, "1080i60  16:9 1920x1080i@119.88/120Hz"},
   {47, "720p120  16:9 1280x720p@119.88/120Hz"},
   {48, "480p119   4:3 720x480p@119.88/120Hz"},
   {49, "480p119H 16:9 720x480p@119.88/120Hz"},
   {50, "480i59    4:3 720(1440)x480i@119.88/120Hz"},
   {51, "480i59H  16:9 720(1440)x480i@119.88/120Hz"},
   {52, "576p200   4:3 720x576p@200Hz"},
   {53, "576p200H 16:9 720x576p@200Hz"},
   {54, "576i100   4:3 720(1440)x576i@200Hz"},
   {55, "576i100H 16:9 720(1440)x576i@200Hz"},
   {56, "480p239   4:3 720x480p@239.76/240Hz"},
   {57, "480p239H 16:9 720x480p@239.76/240Hz"},
   {58, "480i119   4:3 720(1440)x480i@239.76/240Hz"},
   {59, "480i119H 16:9 720(1440)x480i@239.76/240Hz"},
   {60, "720p24   16:9 1280x720p@23.98/24Hz"},
   {61, "720p25   16:9 1280x720p@25Hz"},
   {62, "720p30   16:9 1280x720p@29.97/30Hz"},
   {63, "1080p120 16:9 1920x1080p@119.88/120Hz"}
};

const vmap_t CEA_vidm_map = {
   0, 63,
   vidmd_map
};

//VDB: Video Data Block
const wxString cea_vdb_cl::Name = _("VDB: Video Data Block");
const wxString cea_vdb_cl::Desc = _("Video Data Block contains one or more 1-byte "
"Short Video Descriptors (SVD). Each SVD contains index number from supported resolutions set "
"defined in CEA/EIA standard, and a \"native\" resolution flag");
const size_t cea_vdb_cl::fcount = 2;

rcode cea_vdb_cl::init(uint8_t* inst, uint32_t orflags) {
   rcode retU;
   bhdr_t  bhdr;
   uint    blen;

   if (inst == NULL) {
      RCD_SET_FAULT(retU);
      return retU;
   }
   bhdr.bval = reinterpret_cast <bhdr_t*> (inst)->bval;
   //chk type
//   bhdr.bval = reinterpret_cast <bhdr_t*> (inst)->bval;
//   if (bhdr.tag.blk_type != BTYPE_VDB) {
//      RCD_SET_FAULT(retU);
//      return retU;
//   }
   //chk len
   blen = bhdr.tag.blk_len;
   if (blen % sizeof(svd_t) != 0) {
      RCD_SET_FAULT(retU);
      return retU;
   }
   blen      = blen / sizeof(svd_t);
   type_id   = ID_VDB;
   GroupName = &Name;
   GroupDesc = &Desc;

   RCD_SET_OK(retU);
   svd_t* pgrp_inst = reinterpret_cast <svd_t*> (inst+1);
   subgroups.Alloc(blen);
   //create and init sub-groups: SVD: Short Video Descriptor
   for (uint idx=0; idx<blen; idx++) {
      edi_grp_cl* pgrp = new cea_svd_cl;
      if (pgrp == NULL) {
         RCD_SET_FAULT(retU);
         break;
      }
      retU = pgrp->init(reinterpret_cast <uint8_t*> (pgrp_inst), 0);
      if (! RCD_IS_OK(retU)) break;
      pgrp->setRelOffs(reinterpret_cast <uint8_t*> (pgrp_inst) - inst);
      pgrp->setAbsOffs(abs_offs + pgrp->getRelOffs() );

      subgroups.Add(pgrp);
      pgrp_inst += 1;
   }
   if (! RCD_IS_OK(retU)) {
      subgroups.Clear();
      return retU;
   }

   retU = init_fields(&CEA_BlkHdr_fields[0], inst, fcount, orflags);
   return retU;
}

//VDB: Video Data Block ->
//SVD: Short Video Descriptor
const wxString cea_svd_cl::Name = _("SVD: Short Video Descriptor");
const wxString cea_svd_cl::Desc = _("Short Video Descriptor: "
"contains index number of supported resolutions set "
"defined in CEA/EIA standard, and a \"native\" resolution flag");
const size_t cea_svd_cl::fcount = 2;

const edi_field_t cea_svd_cl::fields[] = {
   {&EDID_cl::BitF8Val, &CEA_vidm_map, 0, 0, 6, EF_BFLD|EF_INT|EF_VS, 0, 0x7F, "vid_mode_idx",
   "Video mode index referencing a table of standard resolutions/timings defined in CEA/EIA-861." },
   {&EDID_cl::BitVal, NULL, 0, 7, 1, EF_BIT, 0, 0, "Native",
   "1= this is \"native\" resolution." }
};

rcode cea_svd_cl::init(uint8_t* inst, uint32_t orflags) {
   rcode retU;

   if (inst == NULL) {
      RCD_SET_FAULT(retU);
      return retU;
   }
   type_id   = ID_SVD;
   GroupName = &Name;
   GroupDesc = &Desc;

   retU = init_fields(&fields[0], inst, fcount, orflags);
   return retU;
}


//VSD: Vendor Specific Data Block: handlers
rcode EDID_cl::VSD_ltncy(uint op, wxString& sval, uint& ival, edi_dynfld_t* field) {
   rcode     retU;
   uint8_t* inst;

   inst = getValPtr(field);
   if (inst == NULL) {
      RCD_SET_FAULT(retU);
      return retU;
   }

   if (op == OP_READ) {
      ival = (inst[0] - 1) << 1; //(val-1)*2
      sval << ival;
      RCD_SET_OK(retU);
   } else {
      ulong       tmpv;
      RCD_SET_FAULT(retU);

      if (op == OP_WRSTR) {
         retU = getStrUint(sval, 10, field->minv, field->maxv, tmpv);
         if (! RCD_IS_OK(retU)) return retU;
      } else if (op == OP_WRINT) {
         tmpv = ival;
         RCD_SET_OK(retU);
      }

      tmpv = (tmpv >> 1)+1;
      inst[0] = (tmpv & 0xFF);
   }
   return retU;
}

rcode EDID_cl::VSD_MaxTMDS(uint op, wxString& sval, uint& ival, edi_dynfld_t* field) {
   rcode     retU;
   uint8_t* inst;

   inst = getValPtr(field);
   if (inst == NULL) {
      RCD_SET_FAULT(retU);
      return retU;
   }

   if (op == OP_READ) {
      ival = (inst[0] * 5);
      sval << ival;
      RCD_SET_OK(retU);
   } else {
      ulong       tmpv;
      RCD_SET_FAULT(retU);

      if (op == OP_WRSTR) {
         retU = getStrUint(sval, 10, field->minv, field->maxv, tmpv);
         if (! RCD_IS_OK(retU)) return retU;
      } else if (op == OP_WRINT) {
         tmpv = ival;
         RCD_SET_OK(retU);
      }

      tmpv = (tmpv / 5);
      inst[0] = (tmpv & 0xFF);
   }
   return retU;
}

//VSD: Vendor Specific Data Block
const wxString cea_vsd_cl::Name = _("VSD: Vendor Specific Data Block");
const wxString cea_vsd_cl::Desc = _("Vendor Specific Data Block is required to contain the "
"following fields: \n1. Vendor's IEEE-id 24-bit LE registration number (For HDMI, it is always "
"00-0C-03 -> HDMI Licensing, LLC).\n2. Source Physical Address (16bit LE). The source physical "
"address provides the CEC physical address for upstream CEC devices.\n"
"The rest of fields can be anything the vendor considers worthy of inclusion in this VSD. "
"HDMI 1.3a specifies some requirements for those data and this is what wxEDID uses as reference.");
const size_t cea_vsd_cl::fcount = 20;

const edi_field_t cea_vsd_cl::fields[] = {
   {&EDID_cl::BitF8Val, NULL, offsetof(vsd_t, tag), 0, 5, EF_BFLD|EF_INT|EF_RD, 0, 0x1F, "blk_len",
   CEA_BlkHdr_dscF0 },
   {&EDID_cl::BitF8Val, NULL, offsetof(vsd_t, tag), 5, 3, EF_BFLD|EF_INT|EF_RD, 0, 0x07, "blk_type",
   CEA_BlkHdr_dscF1 },
   {&EDID_cl::ByteStr, NULL, offsetof(vsd_t, ieee_id), 0, 3, EF_STR|EF_HEX|EF_LE|EF_RD, 0, 0xFFFFFF, "ieee_id",
   "IEEE Registration Id LE, 00-0C-03 for HDMI Licensing, LLC." },
   {&EDID_cl::ByteStr, NULL, offsetof(vsd_t, src_phy), 0, 2, EF_STR|EF_HEX|EF_LE|EF_RD, 0, 0xFFFF, "src phy",
   "Source Physical Address (section 8.7 of HDMI 1.3a)." },
   {&EDID_cl::BitVal, NULL, offsetof(vsd_t, sink_feat), 1, 1, EF_BIT, 0, 1, "DVI_dual",
   "DVI Dual Link Operation support." },
   {&EDID_cl::BitVal, NULL, offsetof(vsd_t, sink_feat), 1, 1, EF_BIT, 0, 1, "reserved",
   "Reserved (0)." },
   {&EDID_cl::BitVal, NULL, offsetof(vsd_t, sink_feat), 2, 1, EF_BIT, 0, 1, "reserved",
   "Reserved (0)." },
   {&EDID_cl::BitVal, NULL, offsetof(vsd_t, sink_feat), 3, 1, EF_BIT, 0, 1, "DC_Y444",
   "4:4:4 support in deep color modes." },
   {&EDID_cl::BitVal, NULL, offsetof(vsd_t, sink_feat), 4, 1, EF_BIT, 0, 1, "DC_30bit",
   "10-bit-per-channel deep color support." },
   {&EDID_cl::BitVal, NULL, offsetof(vsd_t, sink_feat), 5, 1, EF_BIT, 0, 1, "DC_36bit",
   "12-bit-per-channel deep color support." },
   {&EDID_cl::BitVal, NULL, offsetof(vsd_t, sink_feat), 6, 1, EF_BIT, 0, 1, "DC_48bit",
   "16-bit-per-channel deep color support." },
   {&EDID_cl::BitVal, NULL, offsetof(vsd_t, sink_feat), 7, 1, EF_BIT, 0, 1, "AI",
   "Supports_AI (needs info from ACP or ISRC packets)." },
   {&EDID_cl::VSD_MaxTMDS, NULL, offsetof(vsd_t, max_tmds), 0, 1, EF_INT|EF_MHZ, 0, 0xFF, "Max_TMDS",
   "Optional: If non-zero: Max_TMDS_Frequency / 5mhz." },
   {&EDID_cl::BitF8Val, NULL, offsetof(vsd_t, ltncy_hdr), 0, 6, EF_BFLD, 0, 0, "reserved",
   "Reserved (0)." },
   {&EDID_cl::BitVal, NULL, offsetof(vsd_t, ltncy_hdr), 6, 1, EF_BIT, 0, 0, "i_latency",
   "If set, 4 interlaced latency fields are present, must be 0 if latency_f (bit 7) is 0" },
   {&EDID_cl::BitVal, NULL, offsetof(vsd_t, ltncy_hdr), 7, 1, EF_BIT, 0, 0, "latency_f",
   "If set, latency fields are present" },
   {&EDID_cl::VSD_ltncy, NULL, offsetof(vsd_t, vid_lat), 0, 1, EF_INT|EF_MLS, 0, 251, "Video Latency",
   "Optional: Video Latency value=1+ms/2 with a max of 251 meaning 500ms." },
   {&EDID_cl::VSD_ltncy, NULL, offsetof(vsd_t, aud_lat), 0, 1, EF_INT|EF_MLS, 0, 251, "Audio Latency",
   "Optional: Audio Latency value=1+ms/2 with a max of 251 meaning 500ms." },
   {&EDID_cl::VSD_ltncy, NULL, offsetof(vsd_t, vid_ilat), 0, 1, EF_INT|EF_MLS, 0, 251, "Video iLatency",
   "Optional: Interlaced Video Latency value=1+ms/2 with a max of 251 meaning 500ms." },
   {&EDID_cl::VSD_ltncy, NULL, offsetof(vsd_t, aud_ilat), 0, 1, EF_INT|EF_MLS, 0, 251, "Audio iLatency",
   "Optional: Interlaced Audio Latency value=1+ms/2 with a max of 251 meaning 500ms." }
};

rcode cea_vsd_cl::init(uint8_t* inst, uint32_t orflags) {
   rcode retU;

   if (inst == NULL) {
      RCD_SET_FAULT(retU);
      return retU;
   }
   type_id   = ID_VSD;
   GroupName = &Name;
   GroupDesc = &Desc;

   retU = init_fields(&fields[0], inst, fcount, orflags);
   return retU;
}

//SAB: Speaker Allocation Data Block
const wxString cea_sab_cl::Name = _("SAB: Speaker Allocation Block");
const wxString cea_sab_cl::Desc = _("Speaker Allocation Data Block (3 bytes). The second and "
"third are Reserved (0), the first contains information about which speakers are present in "
"the display device.");
const size_t cea_sab_cl::fcount = 12;

const edi_field_t cea_sab_cl::fields [] = {
   {&EDID_cl::BitF8Val, NULL, offsetof(sab_t, tag), 0, 5, EF_BFLD|EF_INT|EF_RD, 0, 0x1F, "blk_len",
   CEA_BlkHdr_dscF0 },
   {&EDID_cl::BitF8Val, NULL, offsetof(sab_t, tag), 5, 3, EF_BFLD|EF_INT|EF_RD, 0, 0x07, "blk_type",
   CEA_BlkHdr_dscF1 },
   {&EDID_cl::BitVal, NULL, offsetof(sab_t, spk_flags), 0, 1, EF_BIT, 0, 1, "front_lr",
   "Front Left / Right present" },
   {&EDID_cl::BitVal, NULL, offsetof(sab_t, spk_flags), 1, 1, EF_BIT, 0, 1, "lfe",
   "LFE present" },
   {&EDID_cl::BitVal, NULL, offsetof(sab_t, spk_flags), 2, 1, EF_BIT, 0, 1, "front_ct",
   "Front Center present" },
   {&EDID_cl::BitVal, NULL, offsetof(sab_t, spk_flags), 3, 1, EF_BIT, 0, 1, "rear_lr",
   "Rear Left / Right present" },
   {&EDID_cl::BitVal, NULL, offsetof(sab_t, spk_flags), 4, 1, EF_BIT, 0, 1, "rear_ct",
   "Rear Center present" },
   {&EDID_cl::BitVal, NULL, offsetof(sab_t, spk_flags), 5, 1, EF_BIT, 0, 1, "frontct_lr",
   "Front Left Center / Right Center present" },
   {&EDID_cl::BitVal, NULL, offsetof(sab_t, spk_flags), 6, 1, EF_BIT, 0, 1, "rearct_lr",
   "Rear Left Center / Right Center present" },
   {&EDID_cl::BitVal, NULL, offsetof(sab_t, spk_flags), 7, 1, EF_BIT, 0, 1, "reserved",
   "Reserved (0)" },
   {&EDID_cl::ByteVal, NULL, offsetof(sab_t, resvdbt1), 0, 1, EF_HEX|EF_RD, 0, 0, "reserved",
   "SAB byte2: Reserved (0)" },
   {&EDID_cl::ByteVal, NULL, offsetof(sab_t, resvdbt2), 0, 1, EF_HEX|EF_RD, 0, 0, "reserved",
   "SAB byte3: Reserved (0)" }
};

rcode cea_sab_cl::init(uint8_t* inst, uint32_t orflags) {
   rcode retU;

   if (inst == NULL) {
      RCD_SET_FAULT(retU);
      return retU;
   }
   type_id   = ID_VSD;
   GroupName = &Name;
   GroupDesc = &Desc;

   retU = init_fields(&fields[0], inst, fcount, orflags);
   return retU;
}

