/***************************************************************
 * Name:      EDID_class.h
 * Purpose:   EDID classes and field handlers
 * Author:    Tomasz Pawlak (tomasz.pawlak@wp.eu)
 * Created:   2014-03-18
 * Copyright: Tomasz Pawlak (C) 2014-2018
 * License:   GPLv3
 **************************************************************/

#ifndef EDID_CLASS_H
#define EDID_CLASS_H 1

#include <wx/string.h>
#include <wx/treectrl.h>
#include <wx/menu.h>
#include <wx/dynarray.h>
#include <wx/arrimpl.cpp>

#include "EDID.h"
#include "EDID_CEA.h"

#include "returncode.h"

typedef unsigned int uint;
typedef unsigned long ulong;

enum { //block IDs
   //base
   ID_BED = 0x00000001,
   ID_VID = 0x00000002,
   ID_BDD = 0x00000004,
   ID_SPF = 0x00000008,
   ID_CXY = 0x00000010,
   ID_ETM = 0x00000020,
   ID_STI = 0x00000040,
   ID_DTD = 0x00000080,
   ID_MRL = 0x00000100,
   ID_MSN = 0x00000200,
   ID_MND = 0x00000400,
   ID_WPD = 0x00000800,
   ID_AST = 0x00001000,
   ID_UST = 0x00002000,
   ID_UNK = 0x00004000,
   //cea
   ID_CHD = 0x00010000,
   ID_ADB = 0x00020000,
   ID_VDB = 0x00040000,
   ID_SAD = 0x00080000,
   ID_SVD = 0x00100000,
   ID_VSD = 0x00200000,
};

enum { //EDID field flags
   //byte0: property flags
   EF_PRCNT  = 6,
   EF_PRSHFT = 0,
   EF_PRMASK = 0x3F,
   EF_RD     = 0x00000001, //read-only
   EF_NI     = 0x00000002, //field value can't be provided as int (for write)
   EF_NU     = 0x00000004, //field is not used i.e. it is marked as unused in the EDID
   EF_FGR    = 0x00000008, //force group refresh after editing field with this flag set
   EF_GPD    = 0x00000010, //display group describtion for this field
   EF_VS     = 0x00000020, //use value selector / values map present
   //byte1: val type
   EF_TPCNT  = 7,
   EF_TPSHFT = 8,
   EF_TPMASK = 0x7F,
   EF_BIT    = 0x00000100, //flag field, single bit
   EF_BFLD   = 0x00000200, //bitfield
   EF_INT    = 0x00000400, //uint
   EF_FLT    = 0x00000800, //float
   EF_HEX    = 0x00001000, //display as hex
   EF_STR    = 0x00002000, //text/byte string
   EF_LE     = 0x00004000, //Little Endian byte order : reverse the byte r/w order for strings
   //byte2: val unit
   EF_UNCNT  = 7,
   EF_UNSHFT = 16,
   EF_UNMASK = 0x7F,
   EF_PIX    = 0x00010000, //
   EF_MM     = 0x00020000, //
   EF_CM     = 0x00040000, //
   EF_HZ     = 0x00080000, //
   EF_KHZ    = 0x00100000, //
   EF_MHZ    = 0x00200000, //
   EF_MLS    = 0x00400000  //miliseconds
};

enum { //handler modes
   OP_READ  = 0x01,
   OP_WRSTR = 0x02,
   OP_WRINT = 0x04
};

class EDID_cl;
struct edi_dynfld_s;
//field - handler fn ptr
typedef rcode (EDID_cl::*field_fn)(uint, wxString&, uint&, edi_dynfld_s*);

enum { //EDI blocks
   EDI_BLK_SIZE = 128,
   EDI_BASE_IDX = 0,
   EDI_EXT0_IDX,
   EDI_EXT1_IDX,
   EDI_EXT2_IDX
};

typedef uint8_t ediblk_t[EDI_BLK_SIZE];

typedef struct __attribute__ ((packed)) { //edi_s
   edid_t   base;
   ediblk_t ext0;
   ediblk_t ext1;
   ediblk_t ext2;
} edi_t;

typedef union __attribute__ ((packed)) { //edi_buff_u
   uint8_t  buff[4* sizeof(edi_t)];
   ediblk_t blk[4];
   edi_t    edi;
} edi_buf_t;

//value map : single entry
typedef struct { // vname_map_t
   const uint  val;
   const char* name;
} vname_map_t;

typedef struct { // vmap_t
   const uint         resvd; //reserved 0
   const uint         nval;
   const vname_map_t *vmap;
} vmap_t;


typedef struct { //edi_field_s
   const field_fn  handlerfn;
   const vmap_t*   vmap;
   const uint      offs;    //offset in data struct - generic/multi-instance functions
   const uint      fldoffs; //offset in bits
   const uint      fldsize; //len in bytes/bits -> depends on flags
   const uint32_t  flags;
   const uint      minv;
   const uint      maxv;
   const char*     name;
   const char*     desc;
} edi_field_t;

typedef struct edi_dynfld_s {
         field_fn  handlerfn;
         vmap_t*   vmap;
         wxMenu*   selector;
         uint8_t*  base;    //data positioning (instance) - generic handlers
         uint      offs;    //offset in data struct - generic/multi-instance handlers
         uint      fldoffs; //offset in bits
         uint      fldsize; //len in bytes/bits -> depends on flags
         uint32_t  flags;
         uint      minv;
         uint      maxv;
   const char*     name;
   const char*     desc;
} edi_dynfld_t;

WX_DECLARE_OBJARRAY(edi_dynfld_t*, GpFieldA_t);

class edi_grp_cl;
WX_DECLARE_OBJARRAY(edi_grp_cl*, GroupsA_t);

class edi_grp_cl : public wxTreeItemData {
   protected:
      uint    type_id;
      uint    abs_offs;
      uint    rel_offs;

      rcode   init_fields(const edi_field_t* field_arr, uint8_t* inst, size_t fcount, uint32_t orflags);
      rcode   create_selector(edi_dynfld_t *pfld);
      void    clear_fields();

   public:
      GpFieldA_t      FieldsA;

      const wxString *GroupName;
      const wxString *GroupDesc;

      inline uint     getTypeID() {return type_id;};
      inline uint     getAbsOffs() {return (abs_offs);};
      inline uint     getRelOffs() {return (rel_offs);};
      inline void     setAbsOffs(uint offs) {abs_offs = offs;};
      inline void     setRelOffs(uint offs) {rel_offs = offs;};

      virtual rcode   init(uint8_t* inst, uint32_t orflags) =0;
      virtual uint    getSubGrpCount() {return 0;};
      virtual edi_grp_cl* getSubGroup(uint ) {return NULL;};
      //virtual GroupsA_t*  getSubGroupA() {return NULL;};

      rcode   ChangeInstancePtr();

      edi_grp_cl() : type_id(0), abs_offs(0), rel_offs(0)  {};
      //~edi_grp_cl() {clear_fields();};
      ~edi_grp_cl() {
         clear_fields();
         FieldsA.Clear();
      };
};


class EDID_cl {
   private:
      edi_buf_t   EDID_buff;

   protected:
      uint         num_valid_blocks;
      bool         RDignore;
      bool         ERRignore;
      wxString     tmps;
      static const wxString val_unit_name[];
      static const wxString val_type_name[];
      static const wxString prop_flag_name[];

      edi_grp_cl* ParseDetDsctors(dsctor_u* pdsc, rcode& retU);
      void        del_groups(GroupsA_t& grparr);

      //Common handlers: helpers
      inline uint     calcGroupOffs (void *inst);
      inline uint8_t* getInstancePtr(edi_dynfld_t* field);
      uint8_t*        getValPtr     (edi_dynfld_t* field);

      rcode getStrUint (wxString& sval, int base, uint minv, uint maxv, ulong& val);
      rcode getStrDoub (wxString& sval, const double minv, const double maxv, double& val);
      rcode rdByteStr  (wxString& sval, uint8_t* pstrb, size_t slen);
      rcode rdByteStrLE(wxString& sval, uint8_t* pstrb, size_t slen);
      rcode wrByteStr  (wxString& sval, uint8_t* pstrb, size_t slen);
      rcode wrByteStrLE(wxString& sval, uint8_t* pstrb, size_t slen);

   public:
      GroupsA_t* BlkGroupsA[4];
      GroupsA_t  EDI_BaseGrpA;
      GroupsA_t  EDI_Ext0GrpA;
      GroupsA_t  EDI_Ext1GrpA;
      GroupsA_t  EDI_Ext2GrpA;

      inline  edi_buf_t* getEDID() {return &EDID_buff;};
      inline  void       SetERRignore(bool rd) {ERRignore = rd;};
      inline  bool       GetERRignore() {return ERRignore;};
      inline  void       SetRDignore(bool rd) {RDignore = rd;};
      inline  bool       GetRDignore() {return RDignore;};
      inline  uint       getNumValidBlocks() {return num_valid_blocks;};
      inline  void       ForceNumValidBlocks(uint nblk) {num_valid_blocks = (nblk%5);}; //max 4
      uint    genChksum(uint block);
      bool    VerifyChksum(uint block);
      void    ClearBuffer();
      rcode ParseEDI_Base();
      rcode ParseEDI_CEA();
      rcode ParseCEA_DBC(uint8_t *pinst, bhdr_t bhdr);

      //field flags
      rcode getValUnitName (wxString& sval, const uint32_t flags);
      rcode getValTypeName (wxString& sval, const uint32_t flags);
      rcode getValFlagsName(wxString& sval, const uint32_t flags);

      //Common handlers
      rcode FldPadString(uint op, wxString& sval, uint& ival, edi_dynfld_t* field);
      rcode ByteStr     (uint op, wxString& sval, uint& ival, edi_dynfld_t* field);
      rcode ByteVal     (uint op, wxString& sval, uint& ival, edi_dynfld_t* field);
      rcode BitF8Val    (uint op, wxString& sval, uint& ival, edi_dynfld_t* field);
      rcode BitVal      (uint op, wxString& sval, uint& ival, edi_dynfld_t* field);
      rcode Gamma       (uint op, wxString& sval, uint& ival, edi_dynfld_t* field);

      //Base Data handlers
      rcode MfcId (uint op, wxString& sval, uint& ival, edi_dynfld_t* field);
      rcode ProdSN(uint op, wxString& sval, uint& ival, edi_dynfld_t* field);
      rcode ProdWk(uint op, wxString& sval, uint& ival, edi_dynfld_t* field);
      rcode ProdYr(uint op, wxString& sval, uint& ival, edi_dynfld_t* field);
      //input type : no dedicated handlers
      //basic display description (old) : no dedicated handlers
      //Supported features : no dedicated handlers
      //Chromacity coords
      rcode ChrXY_getWriteVal(uint op, wxString& sval, uint& ival);
      rcode CHredX (uint op, wxString& sval, uint& ival, edi_dynfld_t* field);
      rcode CHredY (uint op, wxString& sval, uint& ival, edi_dynfld_t* field);
      rcode CHgrnX (uint op, wxString& sval, uint& ival, edi_dynfld_t* field);
      rcode CHgrnY (uint op, wxString& sval, uint& ival, edi_dynfld_t* field);
      rcode CHbluX (uint op, wxString& sval, uint& ival, edi_dynfld_t* field);
      rcode CHbluY (uint op, wxString& sval, uint& ival, edi_dynfld_t* field);
      rcode CHwhtX (uint op, wxString& sval, uint& ival, edi_dynfld_t* field);
      rcode CHwhtY (uint op, wxString& sval, uint& ival, edi_dynfld_t* field);
      //Resolution map : no dedicated handlers
      //std timing descriptors
      rcode StdTxres8(uint op, wxString& sval, uint& ival, edi_dynfld_t* field);
      rcode StdTvsync(uint op, wxString& sval, uint& ival, edi_dynfld_t* field);
      //DTD: Detailed Timing Descriptor
      rcode DTD_PixClk(uint op, wxString& sval, uint& ival, edi_dynfld_t* field);
      rcode DTD_HApix (uint op, wxString& sval, uint& ival, edi_dynfld_t* field);
      rcode DTD_HBpix (uint op, wxString& sval, uint& ival, edi_dynfld_t* field);
      rcode DTD_VAlin (uint op, wxString& sval, uint& ival, edi_dynfld_t* field);
      rcode DTD_VBlin (uint op, wxString& sval, uint& ival, edi_dynfld_t* field);
      rcode DTD_HOsync(uint op, wxString& sval, uint& ival, edi_dynfld_t* field);
      rcode DTD_VOsync(uint op, wxString& sval, uint& ival, edi_dynfld_t* field);
      rcode DTD_VsyncW(uint op, wxString& sval, uint& ival, edi_dynfld_t* field);
      rcode DTD_HsyncW(uint op, wxString& sval, uint& ival, edi_dynfld_t* field);
      rcode DTD_Hsize (uint op, wxString& sval, uint& ival, edi_dynfld_t* field);
      rcode DTD_Vsize (uint op, wxString& sval, uint& ival, edi_dynfld_t* field);
      //MRL / WPD handlers
      rcode MRL_Hdr      (uint op, wxString& sval, uint& ival, edi_dynfld_t* field);
      rcode MRL_GTFM     (uint op, wxString& sval, uint& ival, edi_dynfld_t* field);
      rcode MRL_MaxPixClk(uint op, wxString& sval, uint& ival, edi_dynfld_t* field);

      rcode WPD_pad(uint op, wxString& sval, uint& ival, edi_dynfld_t* field);

      //CEA:VSD
      rcode VSD_ltncy  (uint op, wxString& sval, uint& ival, edi_dynfld_t* field);
      rcode VSD_MaxTMDS(uint op, wxString& sval, uint& ival, edi_dynfld_t* field);


      EDID_cl() : num_valid_blocks(0), RDignore(false), ERRignore(false) {
         BlkGroupsA[0] = &EDI_BaseGrpA;
         BlkGroupsA[1] = &EDI_Ext0GrpA;
         BlkGroupsA[2] = &EDI_Ext1GrpA;
         BlkGroupsA[3] = &EDI_Ext2GrpA;
      };

      ~EDID_cl() {
         EDI_BaseGrpA.Clear();
         EDI_Ext0GrpA.Clear();
         EDI_Ext1GrpA.Clear();
         EDI_Ext2GrpA.Clear();
      };
};

uint  EDID_cl::calcGroupOffs(void *inst) {
   return (reinterpret_cast <uint8_t*> (inst) - EDID_buff.buff);
};

uint8_t* EDID_cl::getInstancePtr(edi_dynfld_t* field) {
   if (field == NULL) return NULL; //reinterpret_cast <uint8_t*> (field);
   return field->base;
};

#ifdef EDID_CLASS

//BED: Base EDID data
class edibase_cl : public edi_grp_cl {
   private:
      static const edi_field_t fields[];

   protected:
      static const wxString Name;
      static const wxString Desc;
      static const size_t   fcount;

   public:
      rcode     init(uint8_t* inst, uint32_t orflags);

      edibase_cl() {};
};
//VID: Video Input Descriptor
class vindsc_cl : public edi_grp_cl {
   private:
      static const edi_field_t fields[];

   protected:
      static const wxString Name;
      static const wxString Desc;
      static const size_t   fcount;

   public:
      rcode     init(uint8_t* inst, uint32_t orflags);
};
//BDD: basic display descriptior
class bddcs_cl : public edi_grp_cl {
   private:
      static const edi_field_t fields[];

   protected:
      static const wxString Name;
      static const wxString Desc;
      static const size_t   fcount;

   public:
      rcode     init(uint8_t* inst, uint32_t orflags);
};
//SFT: Supported features class
class sft_cl : public edi_grp_cl {
   private:
      static const edi_field_t fields[];

   protected:
      static const wxString Name;
      static const wxString Desc;
      static const size_t   fcount;

   public:
      rcode     init(uint8_t* inst, uint32_t orflags);
};
//Chromacity coords class
class chromxy_cl : public edi_grp_cl {
   private:
      static const edi_field_t fields[];

   protected:
      static const wxString Name;
      static const wxString Desc;
      static const size_t   fcount;

   public:
      rcode     init(uint8_t* inst, uint32_t orflags);
};
//ETM: Estabilished Timings Map class
class resmap_cl : public edi_grp_cl {
   private:
      static const edi_field_t fields[];

   protected:
      static const wxString Name;
      static const wxString Desc;
      static const size_t   fcount;

   public:
      rcode     init(uint8_t* inst, uint32_t orflags);
};
//STI: Std Timing Information Descriptor class
class sttd_cl : public edi_grp_cl {
   private:
      static const edi_field_t fields[];

   protected:
      static const wxString Name;
      static const wxString Desc;
      static const size_t   fcount;

   public:
      rcode     init(uint8_t* inst, uint32_t orflags);
};
//DTD fields indexes in array
enum {
   DTD_IDX_PIXCLK = 0,
   DTD_IDX_HAPIX,
   DTD_IDX_HBPIX,
   DTD_IDX_VALIN,
   DTD_IDX_VBLIN,
   DTD_IDX_HSOFFS,
   DTD_IDX_HSWIDTH,
   DTD_IDX_VSOFFS,
   DTD_IDX_VSWIDTH,
   DTD_IDX_HSIZE,
   DTD_IDX_VSIZE,
   DTD_IDX_HBORD,
   DTD_IDX_VBORD,
   DTD_IDX_IL2W,
   DTD_IDX_HSTYPE,
   DTD_IDX_VSTYPE,
   DTD_IDX_SYNCTYP,
   DTD_IDX_STEREOMD,
   DTD_IDX_ILACE
};
//DTD : detailed timing descriptor class
class dtd_cl : public edi_grp_cl {
   private:
      static const edi_field_t fields[];

   protected:
      static const wxString Name;
      static const wxString Desc;
      static const size_t   fcount;

   public:
      rcode     init(uint8_t* inst, uint32_t orflags);
};
//MRL : Monitor Range Limits Descriptor class
class mrl_cl : public edi_grp_cl {
   private:
      static const edi_field_t fields[];

   protected:
      static const wxString Name;
      static const wxString Desc;
      static const size_t   fcount;

   public:
      rcode     init(uint8_t* inst, uint32_t orflags);
};
//WPD : White Point Descriptor class
class wpt_cl : public edi_grp_cl {
   private:
      static const edi_field_t fields[];

   protected:
      static const wxString Name;
      static const wxString Desc;
      static const size_t   fcount;

   public:
      rcode     init(uint8_t* inst, uint32_t orflags);
};
//MND: Monitor Name Descriptor class
class mnd_cl : public edi_grp_cl {
   private:
      static const edi_field_t fields[];

   protected:
      static const wxString Name;
      static const wxString Desc;
      static const size_t   fcount;

   public:
      rcode     init(uint8_t* inst, uint32_t orflags);
};
//MSN: Monitor Serial Number Descriptor class
class msn_cl : public edi_grp_cl {
   private:
      static const edi_field_t fields[];

   protected:
      static const wxString Name;
      static const wxString Desc;
      static const size_t   fcount;

   public:
      rcode     init(uint8_t* inst, uint32_t orflags);
};
//UST: UnSpecified Text class
class ust_cl : public edi_grp_cl {
   private:
      static const edi_field_t fields[];

   protected:
      static const wxString Name;
      static const wxString Desc;
      static const size_t   fcount;

   public:
      rcode     init(uint8_t* inst, uint32_t orflags);
};
//AST: Additional Standard Timing identifiers class
class ast_cl : public edi_grp_cl {
   private:
      static const edi_field_t fields[];

   protected:
      static const wxString Name;
      static const wxString Desc;
      static const size_t   fcount;

   public:
      rcode     init(uint8_t* inst, uint32_t orflags);
};
//UNK: Unknown Descriptor class (type != 0xFA-0xFF)
class unk_cl : public edi_grp_cl {
   private:
      static const edi_field_t fields[];

   protected:
      static const wxString Name;
      static const wxString Desc;
      static const size_t   fcount;

   public:
      rcode     init(uint8_t* inst, uint32_t orflags);
};

#endif // EDID_CLASS

//----------------- CEA/EIA-861 extension

//CHD: CEA Header
class cea_hdr_cl : public edi_grp_cl {
   private:
      static const edi_field_t fields[];

   protected:
      static const wxString Name;
      static const wxString Desc;
      static const size_t   fcount;

   public:
      rcode     init(uint8_t* inst, uint32_t orflags);

      cea_hdr_cl() {};
};
//ADB: Audio Data Block
class cea_adb_cl : public edi_grp_cl {
   private:
      GroupsA_t    subgroups; //ADB_SADs

   protected:

      static const wxString Name;
      static const wxString Desc;
      static const size_t   fcount;

   public:
      //GroupsA_t*   getSubGroupA() {return &subgroups;};
      edi_grp_cl*  getSubGroup(uint idx) {return subgroups.Item(idx);};
      uint         getSubGrpCount() {return subgroups.GetCount();};
      rcode      init(uint8_t* inst, uint32_t orflags);

      cea_adb_cl() {};
};
//ADB: Audio Data Block ->
//SAD: Short Audio Descriptor
class cea_sad_cl : public edi_grp_cl {
   private:
      static const edi_field_t fields[];

   protected:
      static const wxString Name;
      static const wxString Desc;
      static const size_t   fcount;

   public:
      rcode     init(uint8_t* inst, uint32_t orflags);

      cea_sad_cl() {};
};

//VDB: Video Data Block
class cea_vdb_cl : public edi_grp_cl {
   private:
      GroupsA_t    subgroups; //ADB_SADs

   protected:

      static const wxString Name;
      static const wxString Desc;
      static const size_t   fcount;

   public:
      //GroupsA_t*   getSubGroupA() {return &subgroups;};
      edi_grp_cl*  getSubGroup(uint idx) {return subgroups.Item(idx);};
      uint         getSubGrpCount() {return subgroups.GetCount();};
      rcode      init(uint8_t* inst, uint32_t orflags);

      cea_vdb_cl() {};
};
//VDB: Video Data Block ->
//SVD: Short Video Descriptor
class cea_svd_cl : public edi_grp_cl {
   private:
      static const edi_field_t fields[];
      //static const svd_idxm_t  vid_dsc[]; obsolete

   protected:
      static const wxString Name;
      static const wxString Desc;
      static const size_t   fcount;

   public:
      rcode     init(uint8_t* inst, uint32_t orflags);

      cea_svd_cl() {};
};
//VSD: Vendor Specific Data Block
class cea_vsd_cl : public edi_grp_cl {
   private:
      static const edi_field_t fields[];

   protected:
      static const wxString Name;
      static const wxString Desc;
      static const size_t   fcount;

   public:
      rcode     init(uint8_t* inst, uint32_t orflags);

      cea_vsd_cl() {};
};
//SAB: Speaker Allocation Data Block
class cea_sab_cl : public edi_grp_cl {
   private:
      static const edi_field_t fields[];

   protected:
      static const wxString Name;
      static const wxString Desc;
      static const size_t   fcount;

   public:
      rcode     init(uint8_t* inst, uint32_t orflags);

      cea_sab_cl() {};
};

#endif /* EDID_CLASS_H */
