/* Declarations for returncode.h

   Copyright (C) 2013-2018 Tomasz Pawlak, e-mail: tomasz.pawlak@wp.eu
   License: GNU Lesser General Public License version 3 (LGPLv3)

   returncode.h v0.8.12 (2018.12.01)

   This program is free software; You can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public License
   as published by the Free Software Foundation; either version 3
   of the License, or (at your option) any later version.

   This file is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this file; If not, see <http://www.gnu.org/licenses/>.
*/

/* Short description:
   Compacted return type:
   By default it provides exit status, unit code and a source line number.
   The structure can also be used to return message type and number,
   true/false booleans or just any user-defined value, just like a normal
   int32 type.
*/

#ifndef RETURNCODE_H
#define RETURNCODE_H 1

#ifdef __cplusplus
   #ifndef __STDC_LIMIT_MACROS
      #define __STDC_LIMIT_MACROS
   #endif
#endif

#include <stdint.h>
#include <stdio.h>
#include <sys/cdefs.h>

typedef unsigned int uint;

typedef union __attribute__ ((__transparent_union__)) {
   int32_t  value;
   struct __attribute__ ((packed)) detail_s {
      int      rcode :4 ;
      uint     scope :12;
      uint16_t data;
   } detail;
} rcode;

#ifndef RCD_SCOPE
   #define RCD_SCOPE ((uint) 0)
#endif

/* default return code types: */
#define RCD_VAL    ((int)   1) /* indicates that rcode contains meaningful data (detail.data) */
#define RCD_OK     ((int)   0) /* return code ok */
#define RCD_F      ((int)  -1) /* return code fault, line number in data field */
#define RCD_FMSG   ((int)  -2) /* return code fault, message number in data field */
/* pseudo bools */
#define RCD_TRUE   RCD_VAL
#define RCD_FALSE  RCD_OK

#define RCD_NODATA ((uint16_t) 0) /* no data for RCD_SET_VAL arg: hword */

#define RCD_SCOPE_MAX ((uint) 0x0FFF)
#define RCD_DATA_MAX  ((uint) 0xFFFF)

/* RCD_SCOPE:
   It's an identifier for particular unit file. Together with the line number
   in the .data field it defines exact location of an event in a whole project
   or in a sub-module of a project.

   Therefore, the scope number must be unique within a project or whithin a module.
   The easiest way to assure this, is to create a shared header file with a list of scopes:

   scopes.h:
    #define scMain     1
    #define scDataBase 2

   then in main.c:

    #include scopes.h
    #define RCD_SCOPE scMain
    #include returncode.h

   database.c:
    #include scopes.h
    #define RCD_SCOPE scDataBase
    #include returncode.h

   Moreover, to assure that the RCD_SCOPE value is the same for the whole
   unit code, the returncode.h should be included before any other header
   which is referencing returncode.h.
*/

/* check pseudo-bools: (true) */
__always_inline
int RCD_IS_TRUE(rcode rc) {
   return ( rc.detail.rcode >= RCD_TRUE );
}

/* check pseudo-bools: (false) */
__always_inline
int RCD_IS_FALSE(rcode rc) {
   return ( rc.detail.rcode <= RCD_FALSE );
}

/* get scope code from rcode */
__always_inline
uint RCD_GET_SCOPE(rcode rc) {
   return ( rc.detail.scope );
}

/* get rcode from rcode */
__always_inline
int RCD_GET_RCD(rcode rc) {
   return ( rc.detail.rcode );
}

/* get data from rcode */
__always_inline
uint16_t RCD_GET_DATA(rcode rc) {
   return ( rc.detail.data );
}

/* return rcode status: error={RCD_F,RCD_FMSG}, ok={RCD_OK,RCD_VAL} */
__always_inline
int RCD_IS_OK(rcode rc) {
   return ( rc.detail.rcode >= RCD_OK );
}

/* create custom return code */
__always_inline
rcode RCD_SET_VAL(uint scope, int rcd, uint16_t hword) {
   rcode rc;
   rc.value = (int32_t) ( (rcd & 0xF) | ((scope & 0xFFF) << 4) | ((hword & 0xFFFF) << 16) );
   return rc;
}

/* create status OK */
#define RCD_SET_OK( __rcode ) __rcode = RCD_SET_VAL( RCD_SCOPE, RCD_OK, __LINE__ )

/* create fault code */
#define RCD_SET_FAULT( __rcode ) __rcode = RCD_SET_VAL( RCD_SCOPE, RCD_F, __LINE__ )

#define RCD_RETURN_FAULT( __rcode ) { \
   RCD_SET_FAULT( __rcode ); \
   return (__rcode); }

#define RCD_RETURN_OK( __rcode ) { \
   RCD_SET_OK( __rcode ); \
   return (__rcode); }

#define RCD_RETURN_TRUE( __rcode ) { \
   __rcode = RCD_SET_VAL( RCD_SCOPE, RCD_TRUE, __LINE__ ); \
   return (__rcode); }

#define RCD_RETURN_FALSE( __rcode ) { \
   __rcode = RCD_SET_VAL( RCD_SCOPE, RCD_FALSE, __LINE__ ); \
   return (__rcode); }

/* declare and init rcode */
#define RCD_CREATE( __rcode_name ) \
   rcode __rcode_name; RCD_SET_OK( __rcode_name );

#define RCD_COND_RETURN( __rcode ) \
   if (! RCD_IS_OK(__rcode)) return __rcode;

/* RCD_PRINT: print the rcode structure */
__always_inline
void RCD_PRINT(rcode rc, FILE* fd) {
   fprintf(fd, "RCODE=%d, scope=%u, data=%u [ 0x%04X ]\n",
      rc.detail.rcode,
      rc.detail.scope,
      rc.detail.data ,
      rc.detail.data );
}


/* RCD_PRINT for all debug levels */
#ifdef DBG_MODE
   #define RCD_PRINT_DBG( __rc, __fd ) \
      RCD_PRINT( __rc, __fd )
#else
   #define RCD_PRINT_DBG(...)
#endif

#endif /* RETURNCODE_H */
